/*
 * LCMeter.c
 *
 * Created: 05.09.2011 13:56:28
 *  Author: Thomas Oltzen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 */ 

#define F_CPU 16000000UL
#define SECOND_FACTOR 0.999424
#define SECOND_ROUND 61

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>      // Untersttzung fr sleep mode
#include <math.h>
#include <avr/eeprom.h>
#include "lcd.h"

//#define F_CPU  20000000UL
//#define SECOND_FACTOR 1.00925435
//#define SECOND_ROUND 77




// -log(0.5)
#define DISCHARGE_FACTOR 0.69314718

// 4*PI*PI*1nF
#define INDUCTOR_FACTOR 0.00000003947841776043

float refCapacitor EEMEM = 0.000000000022;

volatile uint16_t countTimer2;
volatile uint16_t countTimer1;

// The frequenz is based on 1,00925435 seconds and not 1sec
uint32_t lc_frequenz;
volatile uint8_t out_update = 0;




/* Print a float value on the LCD. */
static void printValue (register float number, register char* str)
{
	register uint8_t space ;
	register short cursor = 13;
	char out[17];
	short neg_sign = 0;
	int no; 
	
	//number = floor(number);
	if (number < 0){
		neg_sign =1;
	}
	
	for (space = 0; space < 16 ; space++) out[space]=' ';
	out[16]=0;
	out[15] = str[1];
	out[14] = str[0];
	space = 0;
	out[13]='0';  // will be overridden if the number bigger or equal one. 
	// cursor>= avoid the infinite loop
	if (number < 10 && number >0){
		out[10]='.';
		out[9]='0'+((int)number % 10);
		number -= ((int)number%10);
		no = number * 1000;
		out[11]='0'+((no/100) % 10);
		out[12]='0'+((no/10) % 10);
		out[13]='0'+(no % 10);
	}

	while(number >= 1 && cursor>=0){
		if (space==3){
			space = 0;
			cursor--;
		} 
		
		space++;
		
		out[cursor]='0'+(int)(fmod (number, 10.0));
		
		number /= 10;
		
		cursor--;
	}
	
	if (neg_sign){
		out[cursor]='-';
	}	
	
	lcd_puts(out);
	
}

/* Set the cursor of LCD on the left side of a line. */

void gotoLine(uint8_t line){
	lcd_gotoxy(0,line);
}

/* Timer 2 count the friction of a second. */
ISR(TIMER2_OVF_vect)  
{
	if (countTimer2==SECOND_ROUND){ // Is one second gone. 
		
		lc_frequenz = 0x10000 * countTimer1 + TCNT1;
		TCNT1 =0;
		countTimer1 =0;
		countTimer2 =0;
		out_update =1;
	}
	countTimer2++;
}

/* Count the overflow of external frequence. */ 
ISR(TIMER1_OVF_vect)  
{
	countTimer1++;
}

/* 
 * Start routine.
 */ 
int main(void)
{
	float f;
	float parameter; // Value of inductor or capacitor
	float ref_cap; 
	uint8_t cal=0;
	set_sleep_mode (SLEEP_MODE_IDLE);
	TIMSK2  |= (1<<TOIE2);		// Overflow aktivieren
	TCCR2B = 7; // Prescale 1024
	TIMSK1  |= (1<<TOIE1);		// Overflow aktivieren
	TCCR1B = 6; // External clock source on T1 pin. Clock on falling edge.

	lcd_init(); 
	lcd_set_cursor_control(0, 0);// No Blink & show cursor
	sei();
	DDRB = 0; // All pins are inputs
	PORTB = 0xff; // All pull up resistors are on
	ref_cap = eeprom_read_float(&refCapacitor);
    for(;;){
		sleep_mode(); // Save power
		if (out_update){
			out_update =0;
			// Fehleranpassung, weil die Schleife nicht genau jede Sekunde aufgerufen wird. 
			f = lc_frequenz / SECOND_FACTOR;
			
			if (!(PINB & (1<<PINB0))){
				// C-Modus
				// The button is an opener
				if (!(PINB & (1<<PINB2)) && cal==0){
					// Messsung
					parameter =(1/(f*DISCHARGE_FACTOR*194000)-ref_cap);
					parameter*=1000000000000.0; 
					gotoLine(0);
					printValue(parameter, "pF");	
				} else {
					// Calibration
						
					parameter =1/(f*DISCHARGE_FACTOR*194000);
					ref_cap = parameter;
					eeprom_write_float (&refCapacitor, parameter);
					gotoLine(0);
					parameter*=1000000000000.0; 
					printValue(parameter, "pF");	
					gotoLine(0);
					lcd_puts("Calib.");
					if (PINB & (1<<PINB2)){
						cal = 4;
					}
					if (cal > 0) cal --;
				}				
			}
			if (!(PINB & (1<<PINB1))){
				// Induction Mode
				parameter =1000000.0/(f*f*INDUCTOR_FACTOR);
				gotoLine(0);
				printValue(parameter, "H"); // print out as H
			}
			
			gotoLine(1);
			printValue(f, "Hz");	
		}			
    }
}
